/*
 * crash-manager
 *
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <dirent.h>
#include <fcntl.h>
#include <getopt.h>
#include <gio/gio.h>
#include <iniparser.h>
#include <inttypes.h>
#include <libgen.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/procfs.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <sys/vfs.h>
#include <unistd.h>

#include <pkgmgr-info.h>
#include <tzplatform_config.h>

#undef LOG_TAG
#define LOG_TAG "CRASH_MANAGER"

#include "defs.h"
#include "shared/log.h"
#include "shared/config.h"
#include "shared/spawn.h"
#include "shared/util.h"
#include "so-info.h"
#include "crash-manager.h"

/* Parsing */
#define KEY_MAX              255

#define CRASH_TEMP_SUBDIR    "/temp/"
#define CRASH_PATH_SUBDIR    "/dump/"
#define LIVE_PATH_SUBDIR     "/livedump/"
#define APP_PATH_SUBDIR      "/app/"

#define WAIT_FOR_OPT_TIMEOUT_SEC 60
#define SPACE_REQUIRED_KB        1024

#define MINICOREDUMPER_TIMEOUT_MS  DEFAULT_COMMAND_TIMEOUT_MS
#define LIVEDUMPER_TIMEOUT_MS      DEFAULT_COMMAND_TIMEOUT_MS
#define CRASH_STACK_TIMEOUT_MS     DEFAULT_COMMAND_TIMEOUT_MS
#define ZIP_TIMEOUT_MS             DEFAULT_COMMAND_TIMEOUT_MS

enum {
	RET_EXCEED = 1,
	NUM_EXCEED,
	USAGE_EXCEED
};

struct file_info {
	bool   isdir;
	size_t size;
	time_t mtime;
	char   *path;
};

/* Configuration variables */
config_t config;
static char* crash_dump_path;
static char* crash_temp_path;

bool have_livecoredumper(void)
{
	return access(LIVEDUMPER_BIN_PATH, X_OK) == 0;
}

/* pkgmgrinfo filter list function for getting application ID */
static int appinfo_get_appid_func(pkgmgrinfo_appinfo_h handle,
		void *user_data)
{
	char *str = NULL;

	int ret = pkgmgrinfo_appinfo_get_appid(handle, &str);
	if (ret == PMINFO_R_OK && str)
		(*(char **)user_data) = strdup(str);

	return ret;
}

static bool get_appid(char *exepath, char **appid)
{
	pkgmgrinfo_appinfo_filter_h handle = NULL;
	bool have_appid = false;

	if (pkgmgrinfo_appinfo_filter_create(&handle) != PMINFO_R_OK)
		return false;

	if (pkgmgrinfo_appinfo_filter_add_string(handle, PMINFO_APPINFO_PROP_APP_EXEC, exepath) != PMINFO_R_OK)
		goto out_free;

	int count = 0;
	if (pkgmgrinfo_appinfo_filter_count(handle, &count) || count < 1)
		goto out_free;

	char *aid = NULL;
	if (pkgmgrinfo_appinfo_filter_foreach_appinfo(handle, appinfo_get_appid_func, &aid) != PMINFO_R_OK)
		goto out_free;

	have_appid = aid != NULL;
	if (have_appid)
		*appid = strdup(aid);

out_free:
	pkgmgrinfo_appinfo_filter_destroy(handle);
	return have_appid && *appid != NULL;
}

static bool get_pkgid(char *appid, char **pkgid)
{
	pkgmgrinfo_appinfo_h handle = NULL;

	if (pkgmgrinfo_appinfo_get_appinfo(appid, &handle) != PMINFO_R_OK)
		return false;

	bool have_pkgid = false;
	char *p = NULL;
	if (pkgmgrinfo_appinfo_get_pkgid(handle, &p) != PMINFO_R_OK)
		goto out_free;

	have_pkgid = p != NULL;
	if (have_pkgid)
		*pkgid = strdup(p);

out_free:
	pkgmgrinfo_appinfo_destroy_appinfo(handle);
	return have_pkgid && *pkgid != NULL;
}

static char *get_app_root_path(const char *pkgid)
{
	char *root_path = NULL;
	pkgmgrinfo_pkginfo_h handle;
	int ret = pkgmgrinfo_pkginfo_get_pkginfo(pkgid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;

	ret = pkgmgrinfo_pkginfo_get_root_path(handle, &root_path);
	if (ret != PMINFO_R_OK) {
		_E("Unable to get app root path for %s", pkgid);
		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
		return NULL;
	}

	char *path = root_path ? strdup(root_path) : NULL;
	if (path == NULL && root_path != NULL)
		_E("Failed to get app root path for %s. strdup() error: %m", pkgid);

	pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return path;
}

static bool set_appinfo(char *exepath, char **appid, char **pkgid)
{
	char *a = NULL, *p = NULL;

	if (get_appid(exepath, &a)) {
		*appid = a;
		*pkgid = get_pkgid(a, &p) ? p : strdup(a);
		return true;
	}

	char *bn = basename(exepath);
	*appid = strdup(bn);
	*pkgid = strdup(bn);
	return false;
}

static int check_disk_available(const char *path, int check_size)
{
	struct statfs lstatfs;

	if (!path)
		return false;

	if (statfs(path, &lstatfs) < 0) {
		_E("Error when checking free space (%d): %m", errno);
		return false;
	}
	int avail_size = (int)(lstatfs.f_bavail * (lstatfs.f_bsize / 1024));

	if (check_size > avail_size) {
		_E("Available disk space (%d) is less than (%d)", avail_size, check_size);
		return false;
	}

	return true;
}

static int prepare_paths(struct crash_info* cinfo)
{
	int tmp_len;
	const char *crash_subdir = cinfo->livedump ? LIVE_PATH_SUBDIR : CRASH_PATH_SUBDIR;

	if (cinfo->output_path) {
		free(config.crash_root_path);
		config.crash_root_path = strdup(cinfo->output_path);
		if (!config.crash_root_path) {
			_E("strdup() error: %m");
			return 0;
		}
	}

	tmp_len = strlen(config.crash_root_path) + strlen(crash_subdir);
	crash_dump_path = (char*)malloc(tmp_len + 1);
	if (crash_dump_path == NULL) {
		_E("Couldn't allocate memory for crash_dump_path: %m\n");
		return 0;
	}
	snprintf(crash_dump_path, tmp_len + 1, "%s%s", config.crash_root_path, crash_subdir);

	tmp_len = strlen(config.crash_root_path) + strlen(CRASH_TEMP_SUBDIR);
	crash_temp_path = (char*)malloc(tmp_len + 1);
	if (crash_temp_path == NULL) {
		_E("Couldn't allocate memory for crash_temp_path: %m\n");
		return 0;
	}
	snprintf(crash_temp_path, tmp_len + 1, "%s%s", config.crash_root_path, CRASH_TEMP_SUBDIR);
	return 1;
}

static void unlock_dir(int fd)
{
	if (flock(fd, LOCK_UN) < 0)
		_E("Failed to unlock file descriptor: %m");
	close(fd);
}

static bool make_dump_dir(void)
{
	const char *dirs[] = {crash_dump_path, crash_temp_path};

	for (size_t i = 0; i < ARRAY_SIZE(dirs); i++) {
		const char *dirname = dirs[i];

		int r = mkdir(dirname, 0775);

		if (r < 0 && errno != EEXIST) {
			_E("Unable to create directory %s: %m", dirname);
			return false;
		}

		chmod(dirname, DEFAULT_CRASH_DIR_PERM); // Fixup permissions for directories created with bad umask

		if (r >= 0)
			continue;

		struct stat st = {0};
		r = stat(dirname, &st);
		bool isdir = !!(st.st_mode & S_IFDIR);

		if (!r && isdir)
			continue;
		else if (!r && !isdir)
			errno = ENOTDIR;

		_E("Failure while trying to ensure %s exists and it is directory: %m", dirname);
		return false;
	}

	return true;
}

static bool get_cmd_info(struct crash_info *cinfo)
{
	assert(cinfo);

	char buf[PATH_MAX];

	if (!read_proc_file(cinfo->pid_info, "cmdline", buf, sizeof(buf), NULL))
		goto err;

	cinfo->cmd_line = strdup(buf);
	if (!cinfo->cmd_line)
		return false;

	if (!read_proc_file(cinfo->tid_info, "comm", buf, sizeof(buf), filter_drop_trailing_whitespace))
		goto err;

	cinfo->comm = strdup(buf);
	if (!cinfo->comm)
		goto err;

	if (!get_exe_path(cinfo->pid_info, buf, sizeof(buf)))
		goto err;

	cinfo->cmd_path = strdup(buf);
	if (!cinfo->cmd_path)
		goto err;

	return true;

err:
	free(cinfo->cmd_line);
	cinfo->cmd_line = NULL;
	free(cinfo->comm);
	cinfo->comm = NULL;
	return false;
}

static int set_prstatus(struct crash_info *cinfo)
{
	int ret;
	char prstatus_name[NAME_MAX+1];

	ret = snprintf(prstatus_name, NAME_MAX, "/%d.prstatus", cinfo->pid_info);
	if (ret < 0) {
		_E("Failed to snprintf for prstatus path");
		goto close_fd;
	}

	cinfo->prstatus_fd = shm_open(prstatus_name, O_RDWR | O_CREAT, 0600);
	if (cinfo->prstatus_fd < 0) {
		_E("shm_open: %m");
		goto close_fd;
	}

	ret = shm_unlink(prstatus_name);
	if (ret < 0) {
		_E("shm_unlink: %m");
		goto close_fd;
	}

	ret = fcntl(cinfo->prstatus_fd, F_GETFD);
	if (ret < 0) {
		_E("fcntl(): %m");
		goto close_fd;
	}

	ret = fcntl(cinfo->prstatus_fd, F_SETFD, ret & ~FD_CLOEXEC);
	if (ret < 0) {
		_E("fcntl(): %m");
		goto close_fd;
	}

	ret = ftruncate(cinfo->prstatus_fd, sizeof(struct elf_prstatus));
	if (ret < 0) {
		_E("ftruncate(): %m");
		goto close_fd;
	}

	return 0;

close_fd:
	if (cinfo->prstatus_fd >= 0)
		close(cinfo->prstatus_fd);
	return -1;
}

void get_proc_name(const char *cmd_line, char *buff, size_t buff_len)
{
	assert(buff != NULL);
	char *space = strchr(cmd_line, ' ');
	size_t name_len;

	if (space == NULL)
		name_len = strlen(cmd_line);
	else
		name_len = space - cmd_line;

	if (name_len > buff_len - 1)
		name_len = buff_len - 1;

	strncpy(buff, cmd_line, name_len);

	buff[name_len] = '\0';
}

static void set_crash_info_defaults(struct crash_info *cinfo)
{
	if (cinfo->livedump) {
		cinfo->sig_info = cinfo->kill ? SIGKILL : 0;
	} else if (cinfo->tid_info == -1)
		cinfo->tid_info = find_crash_tid(cinfo->pid_info);

	if (cinfo->tid_info == -1) {
		_W("TID not known. Assuming TID = PID (%d).", cinfo->pid_info);
		cinfo->tid_info = cinfo->pid_info;
	}
}

static int dump_filter(const struct dirent *de)
{
	assert(de);

	if (de->d_name[0] == '.')
		return 0;
	return 1;
}

/* We treat all errors as if the file was locked. This function is used to check
 * if we can remove the temporary directory. We don't want to remove that
 * directory if an error has occurred, because the deletion will probably also
 * fail */
static bool is_locked(const char *dir_name)
{
	assert(dir_name);

	char lock_file[PATH_MAX];
	if (snprintf(lock_file, sizeof(lock_file), "%s", dir_name) == -1) {
		_E("Unable to check locking status for %s: %m", dir_name);
		return true;
	}

	int fd = open(lock_file, O_RDONLY | O_DIRECTORY);
	if (fd == -1) {
		_E("Unable to open file %s to check if it's blocked: %m", lock_file);
		return true;
	}

	bool result = flock(fd, LOCK_EX | LOCK_NB) == -1;

	close(fd);
	return result;
}

static bool clean_temp(const char *temp_dir)
{
	bool result = true;
	assert(temp_dir);

	struct dirent **scan_list = NULL;
	int scan_num;

	if ((scan_num = scandir(temp_dir, &scan_list, &dump_filter, NULL)) < 0)
		return false;

	for (int i = 0; i < scan_num; i++) {
		char dir_name[PATH_MAX];

		if (snprintf(dir_name, sizeof(dir_name), "%s/%s", temp_dir, scan_list[i]->d_name) == -1) {
			_E("Unable to clean temp for %s: %m", dir_name);
			result = false;
			continue;
		}

		if (is_locked(dir_name))
			_D("Temporary directory %s is locked", dir_name);
		else if (!remove_dir(dir_name, true))
			_W("Can not remove temporary directory: %s", dir_name);
		else
			_D("Temporary directory %s removed", dir_name);
	}

	return result;
}

/* Note: caller of this function is responsible for cleaning up the cinfo on failure */
bool set_crash_info(struct crash_info *cinfo)
{
	set_crash_info_defaults(cinfo);

	if (!get_cmd_info(cinfo)) {
		_E("Failed to get command info");
		return false;
	}

	char date[16];
	struct tm loc_tm;
	cinfo->time_info = time(NULL);
	localtime_r(&cinfo->time_info, &loc_tm);
	strftime(date, sizeof(date), "%Y%m%d%H%M%S", &loc_tm);

	if (asprintf(&cinfo->temp_dir, "%s/crash.XXXXXX", crash_temp_path) == -1)
		goto out_oom;

	if (mkdtemp(cinfo->temp_dir) == NULL || access(cinfo->temp_dir, F_OK)) {
		_E("Failed to create temporary directory %s: %m", cinfo->temp_dir);
		return false;
	}

	char proc_name[PATH_MAX];
	get_proc_name(cinfo->cmd_line, proc_name, sizeof(proc_name));
	const char *suffix = config.report_type >= REP_TYPE_FULL ? (config.allow_zip ? ".zip" : "") : ".info";
	bool is_app = set_appinfo(cinfo->cmd_line, &cinfo->appid, &cinfo->pkgid);
	if (!cinfo->appid || !cinfo->pkgid)
		goto out_oom;

	if (-1 == asprintf(&cinfo->name,        "%s_%d_%s",       is_app ? cinfo->pkgid : basename(proc_name), cinfo->pid_info, date)
	 || -1 == asprintf(&cinfo->pfx,         "%s/%s",          cinfo->temp_dir, cinfo->name)
	 || -1 == asprintf(&cinfo->info_path,   "%s/%s.info",     cinfo->pfx, cinfo->name)
	 || -1 == asprintf(&cinfo->core_path,   "%s/%s.coredump", cinfo->pfx, cinfo->name)
	 || -1 == asprintf(&cinfo->log_path,    "%s/%s.log",      cinfo->pfx, cinfo->name)
	 || -1 == asprintf(&cinfo->zip_path,    "%s/report.zip",  cinfo->temp_dir)
	 || -1 == asprintf(&cinfo->result_path, "%s/%s%s",        crash_dump_path, cinfo->name, suffix))
		goto out_oom;

	if (is_app)
		cinfo->app_root_path = get_app_root_path(cinfo->pkgid);

	if (mkdir(cinfo->pfx, 0775) < 0) {
		_E("Failed to mkdir %s: %m", cinfo->pfx);
		goto out_rm_temp;
	}

	if (set_prstatus(cinfo) < 0)
		goto out_rm_temp;

	return true;

out_rm_temp:
	unlock_dir(cinfo->lock_fd);
	remove_dir(cinfo->temp_dir, true);
	clean_temp(crash_temp_path);
	return false;

out_oom:
	_E("Out of memory");
	return false;
}

static void launch_crash_popup(struct crash_info *cinfo)
{
	assert(cinfo);

	char *av[] = { CRASH_POPUP_BIN_PATH,
			"--cmdline", cinfo->cmd_line,
			"--cmdpath", cinfo->cmd_path,
			NULL };

	spawn_param_s param = { .fn = spawn_nullstdfds };
	spawn(av, NULL, &param, NULL, NULL);
}

static bool dump_system_state(const struct crash_info *cinfo, pid_t *pid)
{
	char *av[] = {DUMP_SYSTEMSTATE_BIN_PATH, "-d", "-k", "-j", "-p", "-e", "-f", cinfo->log_path, "-b", NULL};
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	return spawn(av, NULL, &param, pid, NULL);
}

static void save_so_info(const struct crash_info *cinfo)
{
	char maps_path[PATH_MAX];
	char so_info_path[PATH_MAX];
	snprintf(maps_path, sizeof(maps_path), "/proc/%d/maps",
		 cinfo->pid_info);

	snprintf(so_info_path, sizeof(so_info_path),
			 "%s/%s.%s", cinfo->pfx, cinfo->name, "so_info");

	get_and_save_so_info(maps_path, so_info_path);
}

/* remove a file whose name begins with 'beggining_of_name'. This is needed
 * when we don't want to include coredump in report, but we only know the
 * beginning of the coredump file name. */
static bool remove_file_in_dir(const char *directory, const char *beginning_of_name)
{
	bool ret = false;
	int errno_unlink = 0;

	DIR *dir = opendir(directory);
	if (dir == NULL)
		goto end;

	int dir_fd = dirfd(dir);
	if (dir_fd == -1) {
		closedir(dir);
		goto end;
	}

	struct dirent* file_info;
	while ((file_info = readdir(dir)) != NULL) {
		if (strstr(file_info->d_name, beginning_of_name) == file_info->d_name) {
			ret = unlinkat(dir_fd, file_info->d_name, 0) != -1;
			errno_unlink = errno;
			break;
		}
	}
	closedir(dir);
	errno = errno_unlink; /* for %m */
end:
	return ret;
}

// These macros are used in functions below
#define SNPRINTF_OR_EXIT_W(name, format, member) do { if (snprintf(name##_str, sizeof(name##_str), format, cinfo->member) < 0) goto out; } while (0)
#define SNPRINTF_OR_EXIT(name, format) SNPRINTF_OR_EXIT_W(name, format, name##_info)

static bool extra_script(const struct crash_info *cinfo, pid_t *pid)
{
	if (!config.extra_script)
		return false;

	char pid_str[11];
	SNPRINTF_OR_EXIT(pid, "%d");

	char *av[] = { config.extra_script, cinfo->pfx, pid_str, NULL };
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	return spawn(av, NULL, &param, pid, NULL);

out:
	return false;
}

static void launch_dbus_notify(struct crash_info *cinfo)
{
	assert(cinfo);

	char pid_str[11], tid_str[11], sig_str[11];
	char *prstatus_fd_str = NULL;

	if (asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory: %m");
		return;
	}

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(tid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");

	char *legacy_notification_str = config.legacy_notification ? "--legacy-sig" : "";

	char *av[] = { CRASH_NOTIFY_BIN_PATH,
			"--cmdline", cinfo->cmd_line,
			"--cmdpath", cinfo->cmd_path,
			"--pid", pid_str,
			"--tid", tid_str,
			"--appid", cinfo->appid,
			"--pkgid", cinfo->pkgid,
			"--reportpath", cinfo->result_path,
			"--prstatus_fd", prstatus_fd_str,
			"--signal", sig_str,
			"--tid-comm", cinfo->comm,
			legacy_notification_str,
			NULL };

	spawn(av, NULL, NULL, NULL, NULL);
out:
	free(prstatus_fd_str);
}

static bool execute_minicoredump(struct crash_info *cinfo, int *exit_code)
{
	char *coredump_name = NULL;
	char *prstatus_fd_str = NULL;
	bool is_ok = false;

	if (asprintf(&coredump_name, "%s.coredump", cinfo->name) == -1
	    || asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory");
		goto out;
	}

	char pid_str[11], uid_str[11], gid_str[11], sig_str[11], time_str[11];

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(uid, "%d");
	SNPRINTF_OR_EXIT(gid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");
	SNPRINTF_OR_EXIT(time, "%ld");

	const char *without_core_str = config.dump_core ? NULL : "-s";

	char proc_name[PATH_MAX];
	get_proc_name(cinfo->cmd_line, proc_name, sizeof(proc_name));

	/* Execute minicoredumper */
	char *args[] = {
			MINICOREDUMPER_BIN_PATH,   // minicoredumper filename path
			pid_str,                   // %p - pid
			uid_str,                   // %u - UID
			gid_str,                   // %g - GID
			sig_str,                   // %s - number of signal
			time_str,                  // %t - time of dump
			"localhost",               // %h - hostname
			basename(proc_name), // %e - exe name (need for result filename)
			MINICOREDUMPER_CONFIG_PATH, // config file
			"-d",
			cinfo->pfx,            // temp dir
			"-o",
			coredump_name,             // coredump filename
			"-P",
			prstatus_fd_str,
			(char *)without_core_str,  // with or without core
			NULL
			};

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	is_ok = spawn_wait(args, NULL, &param, MINICOREDUMPER_TIMEOUT_MS, exit_code);

	/* Minicoredumper must be executed to dump at least PRSTATUS for
	   other tools, coredump, however, might have been disabled. */
	if (!config.dump_core && file_exists_in_dir(cinfo->pfx, coredump_name)) {
		_E("Saving core disabled but coredump found (minicoredumper bug?). Removing coredump %s/%s", cinfo->pfx, coredump_name);
		if (!remove_file_in_dir(cinfo->pfx, coredump_name))
			_E("Removing coredump %s/%s failed: %m", cinfo->pfx, coredump_name);
	}

out:
	free(coredump_name);
	free(prstatus_fd_str);

	return is_ok;
}

static bool execute_livedumper(const struct crash_info *cinfo, int *exit_code)
{
	char *coredump_path = NULL;
	char *prstatus_fd_str = NULL;
	bool is_ok = false;
	char pid_str[11];

	if (asprintf(&coredump_path, "%s/%s.coredump", cinfo->pfx, cinfo->name) == -1 ||
	    asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory");
		goto out;
	}

	SNPRINTF_OR_EXIT(pid, "%d");

	/* Execute livedumper */
	char *args[] = {
			LIVEDUMPER_BIN_PATH,    // livedumper filename path
			"-m",                   // minilivecoredump
			"-P", prstatus_fd_str,
			"-f", coredump_path,
			pid_str,                // %p - pid
			NULL
			};

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	is_ok = spawn_wait(args, NULL, &param, LIVEDUMPER_TIMEOUT_MS, exit_code);
out:
	free(prstatus_fd_str);
	free(coredump_path);
	return is_ok;
}

static bool execute_crash_stack(const struct crash_info *cinfo, int *exit_code)
{
	char pid_str[11], tid_str[11], sig_str[11], prstatus_fd_str[11];
	bool is_ok = false;

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(tid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");
	SNPRINTF_OR_EXIT_W(prstatus_fd, "%d", prstatus_fd);

	/* Execute crash-stack */
	char *args[] = { CRASH_STACK_BIN_PATH,
			 "--pid", pid_str,
			 "--tid", tid_str,
			 "--sig", sig_str,
			 "--prstatus_fd", prstatus_fd_str,
			 NULL };

	int fd = open(cinfo->info_path, O_WRONLY | O_CREAT, 0600);
	if (fd < 0) {
		_E("open %s error: %m", cinfo->info_path);
		return false;
	}

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = fd };
	is_ok = spawn_wait(args, NULL, &param, CRASH_STACK_TIMEOUT_MS, exit_code);
	close(fd);

out:
	return is_ok;
}

#undef SNPRINTF_OR_EXIT
#undef SNPRINTF_OR_EXIT_W

static bool kill_pid(const pid_t pid, int sig, bool wait)
{
	if (kill(pid, sig) == -1) {
		_E("kill sig %d error: %m\n", sig);
		return false;
	}

	if (wait) {
		if (waitpid(pid, NULL, 0) == -1) {
			_E("waitpid error: %m");
			return false;
		}
	}
	return true;
}

static bool process_stop(const pid_t pid)
{
	_D("stop process %d", pid);
	return kill_pid(pid, SIGSTOP, false);
}

static bool process_continue(const pid_t pid)
{
	_D("continue process %d", pid);
	return kill_pid(pid, SIGCONT, false);
}

static bool copy_application_data(struct crash_info *cinfo, const char *filename)
{
	bool is_ok = false;
	char *src_filepath = NULL;
	char *dst_filepath = NULL;
	char *dst_dirpath = NULL;

	if (cinfo->app_root_path &&
		(asprintf(&src_filepath, "%s/%s", cinfo->app_root_path, filename) == -1
		|| asprintf(&dst_filepath, "%s%s%s", cinfo->pfx, APP_PATH_SUBDIR, filename) == -1)) {
		_E("Failed to asprintf for %s file in %s path", filename,
				!src_filepath ? cinfo->app_root_path : cinfo->pfx);
		goto out;
	}

	if (!src_filepath || access(src_filepath, F_OK))
		goto out;

	dst_dirpath = strdup(dst_filepath);
	char *dst_dirname = dirname(dst_dirpath);
	if (dst_dirpath == NULL || dst_dirname == NULL) {
		_E("Out of memory");
		goto out;
	}

	if (mkdir(dst_dirname, 0775) < 0 && errno != EEXIST) {
		_E("Failed to mkdir %s: %m", dst_dirname);
		(void)check_disk_available(cinfo->pfx, SPACE_REQUIRED_KB);
		goto out;
	}

	if (copy_file(dst_filepath, src_filepath)) {
		_E("Cannot copy %s file to: %s", src_filepath, dst_filepath);
		(void)check_disk_available(cinfo->app_root_path, SPACE_REQUIRED_KB);
	}
	else
		is_ok = true;
out:
	free(src_filepath);
	free(dst_filepath);
	free(dst_dirpath);

	return is_ok;
}

static void copy_application_xmls(struct crash_info *cinfo)
{
	const char * const app_config_paths[] = {
			tzplatform_getenv(TZ_SYS_RW_PACKAGES),
			tzplatform_getenv(TZ_SYS_RO_PACKAGES)
	};
	char *src_filepath = NULL;
	char *dst_filepath = NULL;
	char *dst_dirpath = NULL;
	char maps_path[PATH_MAX] = {'\0'};

	if (asprintf(&dst_dirpath, "%s%s", cinfo->pfx, APP_PATH_SUBDIR) == -1) {
		_E("Failed to asprintf for dst dir: %s%s", cinfo->pfx, APP_PATH_SUBDIR);
		return;
	}
	if (mkdir(dst_dirpath, 0775) < 0 && errno != EEXIST) {
		_E("Failed to mkdir %s (%m)", dst_dirpath);
		(void)check_disk_available(cinfo->pfx, SPACE_REQUIRED_KB);
		goto out;
	}
	if (snprintf(maps_path, sizeof(maps_path), "/proc/%d/maps", cinfo->pid_info) <= 0) {
		_E("Failed to snprintf maps path: /proc/%d/maps", cinfo->pid_info);
		goto out;
	}

	GHashTable *app_names = get_app_name_from_map(maps_path);
	GHashTableIter iter;
	gpointer key;
	g_hash_table_iter_init(&iter, app_names);
	while (g_hash_table_iter_next(&iter, &key, NULL)) {
		char *app_name = (char *)key;

		if (asprintf(&dst_filepath, "%s%s.xml", dst_dirpath, app_name) == -1) {
			_E("Failed to asprintf dst file path: %s%s%s.xml",
			   cinfo->pfx, APP_PATH_SUBDIR, app_name);
		} else {
			for (size_t i = 0; i < ARRAY_SIZE(app_config_paths); ++i) {
				if (asprintf(&src_filepath, "%s/%s%s", app_config_paths[i], app_name, ".xml") == -1) {
					_E("Failed to asprintf source file path: %s", src_filepath);
					continue;
				}
				if (file_exists(src_filepath)) {
					if (file_exists(dst_filepath)) {
						_W("File %s.xml already exists in report.", app_name);
					} else {
						if (copy_file(dst_filepath, src_filepath)) {
							_E("Cannot copy file from %s to: %s", src_filepath, dst_filepath);
							(void)check_disk_available(dst_dirpath, SPACE_REQUIRED_KB);
						}
					}
				}
				free(src_filepath);
			}
			free(dst_filepath);
		}
		free(app_name);
	}
	g_hash_table_unref(app_names);
out:
	free(dst_dirpath);
}

static bool execute_crash_modules(struct crash_info *cinfo)
{
	int exit_code = 0;
	if (cinfo->livedump) {
		_I("Starting the livedumper");
		if (!process_stop(cinfo->pid_info))
			return false;
		if (!execute_livedumper(cinfo, &exit_code) || exit_code != 0) {
			_E("Failed to run livedumper - can not continue");
			(void)check_disk_available(cinfo->pfx, SPACE_REQUIRED_KB);
			process_continue(cinfo->pid_info);
			return false;
		}
	} else {
		_I("Starting the minicoredumper");
		if (!execute_minicoredump(cinfo, &exit_code) || exit_code != 0) {
			_E("Failed to run minicoredumper - can not continue");
			(void)check_disk_available(cinfo->pfx, SPACE_REQUIRED_KB);
			return false;
		}
	}

	execute_crash_stack(cinfo, NULL);
	(void)copy_application_data(cinfo, "tizen-manifest.xml"); // manifest is optional because only tizen applications have it
	copy_application_xmls(cinfo);
	if (cinfo->livedump)
		process_continue(cinfo->pid_info);

	return true;
}

static int lock_dir(const char *path, bool block)
{
	assert(path);

	int fd;

	if ((fd = open(path, O_RDONLY | O_DIRECTORY)) == -1) {
		_E("Failed to open %s: %m", path);
		return -1;
	}

	int flock_flags = LOCK_EX;

	if (!block)
		flock_flags |= LOCK_NB;

	if (flock(fd, flock_flags) == -1) {
		_E("Failed to lock %s for exclusive access: %m", path);
		close(fd);
		return -1;
	}

	return fd;
}

static int mtime_cmp(const void *_a, const void *_b)
{
	const struct file_info *a = _a;
	const struct file_info *b = _b;

	if (a->mtime < b->mtime)
		return -1;
	if (a->mtime > b->mtime)
		return 1;
	return 0;
}

static int scan_dump(struct file_info **dump_list, off_t *usage)
{
	struct file_info *temp_list;
	struct dirent **scan_list = NULL;
	struct stat st;
	int i, scan_num, dump_num = 0;
	int fd;

	if ((fd = open(crash_dump_path, O_DIRECTORY)) < 0) {
		_E("Failed to open %s: %m", crash_dump_path);
		return -1;
	}

	scan_num = scandir(crash_dump_path, &scan_list, &dump_filter, NULL);
	if (scan_num < 0) {
		close(fd);
		return -1;
	}

	temp_list = (struct file_info *)calloc(scan_num,
			sizeof(struct file_info));
	if (!temp_list) {
		_E("Failed to calloc for dump list");
		goto exit;
	}

	for (i = 0; i < scan_num; i++) {
		if (fstatat(fd, scan_list[i]->d_name, &st, 0) < 0) {
			_E("Failed to fstatat");
			continue;
		}

		if (asprintf(&(temp_list[dump_num].path), "%s/%s",
					crash_dump_path, scan_list[i]->d_name) < 0) {
			_E("Failed to asprintf");
			continue;
		}

		if (scan_list[i]->d_type == DT_DIR) {
			temp_list[dump_num].isdir = 1;
			temp_list[dump_num].size =
				get_directory_usage(temp_list[dump_num].path);
		} else {
			temp_list[dump_num].isdir = 0;
			temp_list[dump_num].size = st.st_size;
		}
		temp_list[dump_num].mtime = st.st_mtime;
		dump_num++;
	}

	if (dump_num <= 0) {
		free(temp_list);
		goto exit;
	}

	if (dump_num != scan_num) {
		struct file_info *const tmp = (struct file_info *)realloc(temp_list,
				dump_num * sizeof(struct file_info));
		if (!tmp) {
			free(temp_list);
			dump_num = -1;
			goto exit;
		}
		temp_list = tmp;
	}

	qsort(temp_list, dump_num, sizeof(struct file_info), mtime_cmp);

	*usage = 0;
	for (i = 0; i < dump_num; i++) {
		_D("[%d] path: %s(%s), size: %zu kb, mtime: %s",
				i,
				temp_list[i].path,
				temp_list[i].isdir ? "DIR" : "FILE",
				temp_list[i].size / 1024,
				ctime(&(temp_list[i].mtime)));
		*usage += temp_list[i].size;
	}
	*dump_list = temp_list;
exit:
	for (i = 0; i < scan_num; i++)
		free(scan_list[i]);
	free(scan_list);
	close(fd);

	return dump_num;
}

static bool remove_file_info(struct file_info file)
{
	if (file.isdir)
		return remove_dir(file.path, true);
	else
		return unlink(file.path) != -1;
}

static void clean_dump(void)
{
	struct file_info *dump_list = NULL;
	int i, scan_num, dump_num, remove_flag;
	off_t usage = 0;
	time_t cur_time;

	scan_num = scan_dump(&dump_list, &usage);
	if (scan_num <= 0)
		return;

	dump_num = scan_num;

	cur_time = time(NULL);

	for (i = 0; i < scan_num; i++) {
		remove_flag = 0;

		/* Retention time check */
		if (config.max_retention_sec &&
			dump_list[i].mtime > 0 &&
			dump_list[i].mtime + config.max_retention_sec < cur_time)
			remove_flag = RET_EXCEED;
		/* Check the number of dumps */
		else if (config.max_crash_dump &&
			0 < dump_num && config.max_crash_dump < dump_num)
			remove_flag = NUM_EXCEED;
		/* Check the max system use size */
		else if (config.system_max_use &&
			0 < dump_num && config.system_max_use < usage / 1024)
			remove_flag = USAGE_EXCEED;

		switch (remove_flag) {
		case RET_EXCEED:
			_I("Reached the maximum retention time %d, so remove (%s)",
					config.max_retention_sec, dump_list[i].path);
			break;
		case NUM_EXCEED:
			_I("Reached the maximum number of dump %d/%d, so remove (%s)",
					dump_num, config.max_crash_dump,
					dump_list[i].path);
			break;
		case USAGE_EXCEED:
			_I("Reached the maximum disk usage %" PRId64 "/%d kb, so remove (%s)",
					usage / 1024, config.system_max_use,
					dump_list[i].path);
			break;
		default:
			break;
		}

		if (!remove_flag)
			continue;

		if (!remove_file_info(dump_list[i])) {
			_E("Failed to remove %s", dump_list[i].path);
			continue;
		}
		dump_num--;
		usage -= dump_list[i].size;
	}

	/* Check disk free space to keep */
	if (config.system_keep_free &&
			!check_disk_available(config.crash_root_path,
					     config.system_keep_free)) {
		_I("Disk is not available! so set the maximum number of dump to 1");
		config.max_crash_dump = 1;
	}

	for (i = 0; i < dump_num; i++)
		free(dump_list[i].path);
	free(dump_list);
}

static void compress(struct crash_info *cinfo)
{
	char *args[] = {"/bin/zip", "-qyr", cinfo->zip_path, cinfo->name, NULL};
	spawn_param_s param1 = { .fn = spawn_nullstdfds };
	spawn_param_s param0 = { .fn = spawn_chdir, .u.char_ptr = cinfo->temp_dir, .next = &param1 };
	(void)spawn_wait(args, NULL, &param0, ZIP_TIMEOUT_MS, NULL);
}

static bool move_dump_data(const char *from_path, const struct crash_info *cinfo)
{
	int lock_fd;
	bool is_ok = true;

	if ((lock_fd = lock_dir(crash_dump_path, false)) < 0)
		return false;
	if (!rename(from_path, cinfo->result_path)) {
		chmod(cinfo->result_path, DEFAULT_REPORT_PERM);
		clean_dump();
	} else {
		_E("Failed to move %s to %s", from_path, cinfo->result_path);
		(void)check_disk_available(crash_dump_path, SPACE_REQUIRED_KB);
		is_ok = false;
	}
	unlock_dir(lock_fd);

	return is_ok;
}

static int wait_for_opt(unsigned int timeout)
{
	unsigned int count = 0;

	while (!check_disk_available(config.crash_root_path, 0) && count < timeout) {
		log_kmsg("crash-manager: path %s is not available\n", config.crash_root_path);
		sleep(1);
		count++;
	}
	if (count >= timeout) {
		log_kmsg("crash-manager: timeout (%ds) while waiting for %s."
			"Probably /opt is not mounted.\n", timeout, config.crash_root_path);
		return 0;
	}

	return 1;
}

static void free_crash_info(struct crash_info *cinfo)
{
	free(cinfo->cmd_line);
	free(cinfo->cmd_path);
	free(cinfo->comm);
	free(cinfo->temp_dir);
	free(cinfo->result_path);
	free(cinfo->app_root_path);
	free(cinfo->pfx);
	free(cinfo->info_path);
	free(cinfo->core_path);
	free(cinfo->log_path);
	free(cinfo->zip_path);
	free(cinfo->appid);
	free(cinfo->pkgid);
	free(cinfo->executable_path);
}

void crash_info_init(struct crash_info *cinfo)
{
	cinfo->prstatus_fd = -1;
	cinfo->livedump = false;
	cinfo->kill = false;
	cinfo->print_result_path = false;
	cinfo->tid_info = -1;
	cinfo->time_info = 0;
	cinfo->output_path = NULL;
	cinfo->executable_path = NULL;
	cinfo->cmd_line = NULL;
	cinfo->cmd_path = NULL;
	cinfo->comm = NULL;
	cinfo->temp_dir = NULL;
	cinfo->app_root_path = NULL;
	cinfo->pfx = NULL;
	cinfo->result_path = NULL;
	cinfo->info_path = NULL;
	cinfo->core_path = NULL;
	cinfo->log_path = NULL;
	cinfo->zip_path = NULL;
	cinfo->appid = NULL;
	cinfo->pkgid = NULL;
	cinfo->lock_fd = -1;
}

static void release_crashed_process()
{
	/* Release the core pipe as passed by kernel, allowing another
	 * coredump to be handled.
	 *
	 * Due to usage of core_pipe_limit there is limited number of
	 * crash-manager processes that kernel is going to invoke
	 * concurrently.  As the next and last step is a _synchronous_
	 * call to crash-popup we close the descriptor here.
	 *
	 * Note: for VIP processes this will likely cause the system
	 * to reboot without showing popup.
	 */
	close(STDIN_FILENO);
	_I("Released crashed process lock");
}

static bool run(struct crash_info *cinfo)
{
	/* Execute processes in parallel */
	static pid_t dump_state_pid = 0, extra_script_pid = 0;
	int debug_mode = access(DEBUGMODE_PATH, F_OK) == 0;

	if (config.report_type >= REP_TYPE_FULL) {
		/* Exec dump_systemstate */
		if (!dump_system_state(cinfo, &dump_state_pid)) {
			_E("Failed to get system state report");
			return false;
		}

		if (config.extra_script && !extra_script(cinfo, &extra_script_pid))
			_W("Failed to call extra script from config");
	}

	_I("Creating report for pid %d, tid %d, cmdline %s, pkgid %s",
	   cinfo->pid_info, cinfo->tid_info, cinfo->cmd_line, cinfo->pkgid);

	/* Exec crash modules */
	if (!execute_crash_modules(cinfo)) {
		_E("Failed to get basic crash information");
		return false;
	}

	if (!cinfo->livedump && config.release_early)
		release_crashed_process();

	char *temp_report;
	if (config.report_type >= REP_TYPE_FULL) {
		/* Save shared objects info (file names, bulid IDs, rpm package names) */
		if (config.dump_so_info)
			save_so_info(cinfo);
		else
			_I("Not saving .so_info (disabled in configuration)");

		/* Wait misc. pids */
		wait_for_pid(dump_state_pid, NULL);
		if (extra_script_pid > 0)
			wait_for_pid(extra_script_pid, NULL);

		if (config.allow_zip)
			compress(cinfo);

		temp_report = config.allow_zip ? cinfo->zip_path : cinfo->pfx;
	} else
		temp_report = cinfo->info_path;

	if (move_dump_data(temp_report, cinfo))
		_I("Report for pid %d created at %s", cinfo->pid_info, cinfo->result_path);

	if (cinfo->print_result_path)
		printf("REPORT_PATH=%s\n", cinfo->result_path);

	if (!cinfo->livedump) {
		if (!config.release_early)
			release_crashed_process();

		launch_dbus_notify(cinfo);

		/* launch crash-popup only if the .debugmode file exists */
		if (debug_mode)
			launch_crash_popup(cinfo);
	} else if (cinfo->kill) {
		kill_pid(cinfo->pid_info, SIGKILL, false);
	}

	return true;
}

static bool crash_manager_prepare(struct crash_info *cinfo)
{
	if (!config_init(&config, CRASH_MANAGER_CONFIG_PATH))
		return false;

	if (cinfo->executable_path && config_is_path_excluded(&config, cinfo->executable_path))
		return false;

	if (!prepare_paths(cinfo))
		return false;

	if (!wait_for_opt(WAIT_FOR_OPT_TIMEOUT_SEC))
		return false;

	/* Create crash directories */
	if (!make_dump_dir())
		return false;

	if (!set_crash_info(cinfo))
		return false;

	cinfo->lock_fd = lock_dir(cinfo->temp_dir, true);
	if (cinfo->lock_fd == -1)
		return false;

	return true;
}

static void write_dump_reason(const char *reason, const char *base_dir, const char *name)
{
	char *reason_name;

	if (asprintf(&reason_name, "%s.dump_reason", name) == -1) {
		_E("Failed to asprintf for reason_name: %m");
	} else {
		write_to_file(reason, base_dir, reason_name);
		free(reason_name);
	}
}

static void crash_manager_cleanup(struct crash_info *cinfo)
{
	assert(cinfo);
	unlock_dir(cinfo->lock_fd);

	if (!remove_dir(cinfo->temp_dir, true))
		_E("Failed to delete temp directory: %s", cinfo->temp_dir);

	clean_temp(crash_temp_path);
}

void crash_manager_free(struct crash_info *cinfo)
{
	if (cinfo->prstatus_fd >= 0)
		close(cinfo->prstatus_fd);
	free(crash_temp_path);
	free(crash_dump_path);
	config_free(&config);

	free_crash_info(cinfo);
}

bool crash_manager_direct(struct crash_info *cinfo)
{
	if (!crash_manager_prepare(cinfo))
		return false;

	bool result = run(cinfo);
	crash_manager_cleanup(cinfo);

	return result;
}

bool crash_manager_livedump_pid(pid_t pid, const char *dump_reason, char *report_path, size_t report_path_len)
{
	bool result = false;
	struct crash_info cinfo;
	crash_info_init(&cinfo);

	cinfo.livedump = true;
	cinfo.pid_info = pid;

	if (!crash_manager_prepare(&cinfo))
		goto exit;

	write_dump_reason(dump_reason, cinfo.pfx, cinfo.name);

	result = run(&cinfo);
	crash_manager_cleanup(&cinfo);

	if (!result)
		goto exit;

	strncpy(report_path, cinfo.result_path, report_path_len);

	result = true;
exit:
	crash_manager_free(&cinfo);
	_I("Exiting with %s", result ? "success" : "fail");
	return result;
}

