/*
 * This file is a part of dump_systemstate from the crash-worker project.
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// dump_systemstate
#include "dump_systemstate.h"
#include "extras.h"

// crash-worker
#include "shared/spawn.h"
#include "shared/util.h"

// external projects
#include <iniparser.h>

// POSIX
#include <dirent.h>
#include <fcntl.h>

// C
#include <assert.h>
#include <limits.h>
#include <stdbool.h>

static inline void cleanup_dictionary(dictionary **ini)
{
	assert(ini);
	if (!*ini)
		return;

	iniparser_freedict(*ini);
}

enum ini_fields {
	INI_FIELD_TITLE = 0,
	INI_FIELD_PATH,
	INI_FIELD_ARGS,
	INI_FIELD_ENV,
	INI_FIELD_FLAG,
	COUNT_INI_FIELDS,
};

static const char *const INI_KEYS[COUNT_INI_FIELDS] = {
	[INI_FIELD_TITLE] = "title",
	[INI_FIELD_PATH]  = "path",
	[INI_FIELD_ARGS]  = "args",
	[INI_FIELD_ENV]   = "env",
	[INI_FIELD_FLAG]  = "cmdflag",
	// non-string field: "order"
};
static const size_t MAX_INI_KEY_LEN = 7;

struct extra_dump_item {
	// not separate named fields, for convenient iteration
	char *fields[COUNT_INI_FIELDS];
	int order;
};

void cleanup_extra_dump_item(struct extra_dump_item *edi)
{
	for (size_t i = 0; i < ARRAY_SIZE(edi->fields); ++i)
		free(edi->fields[i]);
}

struct extra_items_vector {
	size_t size;
	struct extra_dump_item *data;
};

bool check_cmdflag(const char *const flag, int argc, char **argv)
{
	if (!flag)
		return true;

	for (int i = 1; i < argc; i++) {
		char *p = argv[i];
		if (*p != '-')
			continue;
		for (++p; *p; p++) {
			if (*p == '-')
				break;
			if (*p == flag[0])
				return true;
		}
	}

	return false;
}

int handle_extra_program(int out_fd, struct extra_dump_item *item, int argc, char **argv)
{
	assert(out_fd >= 0);
	assert(item);

	char *const title = item->fields[INI_FIELD_TITLE];
	char *const path  = item->fields[INI_FIELD_PATH];
	char *const args  = item->fields[INI_FIELD_ARGS] ?: "";
	char *const env   = item->fields[INI_FIELD_ENV] ?: "";
	char *const flag  = item->fields[INI_FIELD_FLAG];

	if (!check_cmdflag(flag, argc, argv))
		return 0;

	if (!title || !path) {
		fprintf_fd(out_fd, "\nNo title or path in extra program config");
		return EXIT_CONFERR;
	}

	char *command_line;
	int printed = asprintf(&command_line, "%s %s %s", env, path, args);
	if (printed < 0) {
		fprintf_fd(out_fd, "\nError, out of memory");
		return EXIT_ERR;
	}

	fprintf_fd(out_fd, "\n==== %s (%s)\n", title, command_line);

	/* Both `args` and `env` are single strings (because `iniparser` does not
	 * offer much when it comes to hierarchies and arrays) but `execve` expects
	 * an array of char pointers. Splitting isn't trivial (consider a brutal set
	 * of arguments using " or `) and I don't want to reinvent the wheel so I'm
	 * delegating the splitting to the shell. */
	char *av[] = {"/bin/sh", "-c", command_line, NULL};

	int err;
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = out_fd };
	bool failed = !spawn_wait(av, NULL, &param, DEFAULT_COMMAND_TIMEOUT_MS, &err) || err != 0;

	free(command_line);
	return failed ? EXIT_CMDERR : 0;
}

int handle_extra_file(int out_fd, struct extra_dump_item *item, int argc, char **argv)
{
	assert(out_fd >= 0);
	assert(item);

	char *const title = item->fields[INI_FIELD_TITLE];
	char *const path  = item->fields[INI_FIELD_PATH];
	if (!title || !path) {
		fprintf_fd(out_fd, "\nNo title or path in extra file config");
		return EXIT_CONFERR;
	}

	fprintf_fd(out_fd, "\n==== %s (%s)\n", title, path);
	int ret = dump_file_write_fd(out_fd, (char *)path);
	if (ret < 0) {
		fprintf_fd(out_fd, "Unable to copy file.\n");
		return EXIT_FILEERR;
	}
	return 0;
}

typedef int (*handle_ini_section_t)(int out_fd, struct extra_dump_item *, int argc, char **argv);

static int handle_ini_Nth_section(struct extra_dump_item *item, dictionary *ini, int n)
{
	assert(item);
	assert(ini);
	assert(n >= 0);
	assert(n < iniparser_getnsec(ini));

	char *const secname = iniparser_getsecname(ini, n);
	assert(secname); // can only be NULL if `ini` is NULL or `n` is outta bounds

	const size_t secname_len = strlen(secname);
	char key_buf[secname_len + sizeof ':' + MAX_INI_KEY_LEN + sizeof '\0'];
	memcpy(key_buf, secname, secname_len);
	key_buf[secname_len] = ':';

	int ret = 0;
	char *key_suffix_ptr = key_buf + secname_len + 1;
	for (size_t i = 0; i < ARRAY_SIZE(item->fields); ++i) {
		strcpy(key_suffix_ptr, INI_KEYS[i]);

		char *tmp = iniparser_getstring(ini, key_buf, NULL);
		if (!tmp) {
			item->fields[i] = NULL;
			continue;
		}

		item->fields[i] = strdup(tmp);
		if (!item->fields[i])
			ret |= EXIT_ERR;
	}

	strcpy(key_suffix_ptr, "order");
	char *tmp = iniparser_getstring(ini, key_buf, NULL);
	item->order = tmp ? atoi(tmp) : INT_MAX;

	return ret;
}

static int handle_extra_ini(int out_fd, struct extra_items_vector *eiv, const char *ini_path)
{
	assert(out_fd >= 0);
	assert(eiv);
	assert(ini_path);

	__attribute__((cleanup(cleanup_dictionary))) dictionary *ini = iniparser_load(ini_path);
	if (!ini) {
		fprintf_fd(out_fd, "\nCouldn't parse ini file %s", ini_path);
		return EXIT_CONFERR;
	}

	const int nsec = iniparser_getnsec(ini);
	assert(nsec >= 0); // can only be -1 when ini is NULL

	const size_t prev_size = eiv->size;
	struct extra_dump_item *const temp = realloc(eiv->data, (eiv->size + nsec) * sizeof *eiv->data);
	if (!temp)
		return EXIT_ERR;
	eiv->data = temp;
	eiv->size += nsec;

	int ret = 0;
	for (int i = 0; i < nsec; ++i)
		ret |= handle_ini_Nth_section(eiv->data + prev_size + i, ini, i);
	return ret;
}

static int config_entry_filter(const struct dirent *de)
{
	assert(de);
	return de->d_type == DT_REG && string_ends_with(de->d_name, ".conf");
}

int handle_items_vector(struct extra_items_vector *eiv, int out_fd, handle_ini_section_t handle_ini_section, int argc, char **argv)
{
	inline int cmp(const void *a, const void *b) {
		return ((struct extra_dump_item *)a)->order - ((struct extra_dump_item *)b)->order;
	}
	qsort(eiv->data, eiv->size, sizeof *eiv->data, cmp);

	int ret = 0;
	for (size_t i = 0; i < eiv->size; ++i)
		ret |= handle_ini_section(out_fd, eiv->data + i, argc, argv);
	return ret;
}

void free_items_vector(struct extra_items_vector *eiv)
{
	for (size_t i = 0; i < eiv->size; ++i)
		cleanup_extra_dump_item(eiv->data + i);
	free(eiv->data);
}

int handle_extra_dir(int out_fd, char *dir_path, handle_ini_section_t handle_ini_section, int argc, char **argv)
{
	assert(out_fd >= 0);
	assert(dir_path);
	assert(handle_ini_section);

	const int dir_fd = open(dir_path, O_DIRECTORY | O_RDONLY);
	if (dir_fd < 0) {
		fprintf_fd(out_fd, "\nCouldn't open extras dir: %s %m", dir_path);
		return EXIT_ERR;
	}

	struct dirent **entries;
	int entry_count = scandirat(dir_fd, ".", &entries, config_entry_filter, alphasort);
	if (entry_count < 0) {
		fprintf_fd(out_fd, "\nCouldn't process directory %s: %m", dir_path);
		close(dir_fd);
		return EXIT_ERR;
	}

	__attribute__((cleanup(free_items_vector))) struct extra_items_vector eiv = {
		.size = 0,
		.data = NULL,
	};

	int ret = 0;
	for (int i = 0; i < entry_count; ++i) {
		struct dirent *const de = entries[i];
		char ini_path[PATH_MAX];

		/* In theory this introduces a race condition (somebody could replace
		 * folders between `scandirat` and the file being read) but `iniparser`
		 * does not offer any other way to open a file. Not letting the system
		 * get so brutally pwned is not our responsibility, at any rate. */
		snprintf(ini_path, sizeof ini_path, "%s/%s", dir_path, de->d_name);
		free(de);

		ret |= handle_extra_ini(out_fd, &eiv, ini_path);
	}
	ret |= handle_items_vector(&eiv, out_fd, handle_ini_section, argc, argv);

	free(entries);
	close(dir_fd);
	return ret;
}

